package merlin.wizard;

import merlin.filter.*;
import merlin.util.*;
import java.lang.*;
import java.util.*;
import java.awt.*;

/**
 * The interface of a ElasticWizard. 
 * An ElasticWizard takes a Pattern and 
 * produces a guess of what it looks like, using the technique known as
 * elastic matching discussed e.g. in 
 * <I><A HREF="http://www.testbed.era.ericsson.se/~erasshe/reports.html">
 * Handwriting Recognition using Weighted Elastic Matching</A></I>, by
 * Patrice Scattolin. 
 * Copyright (c) Stefan Hellkvist (see COPYRIGHT)
 * @author Stefan Hellkvist
 * @version 1.0
 */

public class ElasticWizard extends Wizard
{
    private static final boolean DEBUG = false;
    private static final double C = 1000.0; /* 50.0 */
    private static final int NUM_BANKS = 3;
    private static final int THRESHOLD = 49;
    private static final int N_CANDIDATES = 6;
    private static final int QUICK_SAMPLE_RATE = 3;
    private static final int NORMAL_SAMPLE_RATE = 1;

    /**
     * Default constructor
     */
    public ElasticWizard()
	{
	    initPreprocessor();
	    initSamples();
	}

    /**
     * You give the ElasticWizard a pattern and lets him recognize it.
     * You don't need to preprocess the pattern with any filter.
     * The ElasticWizard will do all the necessary filtering.
     * @param p  The pattern to recognize
     * @return The ascii value of the pattern
     */
    public char recognize(Pattern p)
	{
	    int mini = 0, mind, tmpd;
	    if (maxWidth(p) < 15)
		return '.';
	    p = preprocessor.apply(p);

	    byte[][] pointAt = cleanStroke(p);
	    float[] slopeAt = calcSlopes(pointAt);
	    char[] ascii = null;

	    switch(getBank())
	    {
	    case NATURAL: samples = Samples.getNatural();
		ascii = Samples.getNaturalAscii(); break;
	    case NUMERAL: samples = Samples.getNumeral();
		ascii = Samples.getNumeralAscii(); break;
	    case EXTENDED: samples = Samples.getExtended();
		ascii = Samples.getExtendedAscii(); break;
	    };

	    currentUnknownPoint = filter(QUICK_SAMPLE_RATE, pointAt);
	    currentUnknownSlope = filter(QUICK_SAMPLE_RATE, slopeAt);

	    Vector v = candidates();

	    currentUnknownPoint = filter(NORMAL_SAMPLE_RATE, pointAt);
	    currentUnknownSlope = filter(NORMAL_SAMPLE_RATE, slopeAt);

	    mini = ((Integer) v.elementAt(0)).intValue();
	    mind = distance(mini, NORMAL_SAMPLE_RATE); 
	    
	    for (int i = 1; i < v.size(); i++)
	    {
		int index = ((Integer) v.elementAt(i)).intValue();
		tmpd = distance(index, NORMAL_SAMPLE_RATE);
		if (tmpd < mind)
		{
		    mind = tmpd;
		    mini = index;
		}
	    }
	    return ascii[mini];
	}
    
    private int maxWidth(Pattern p)
	{
	    int minx, maxx, miny, maxy;
	    Point p1 = (Point) ((Vector) p.strokes().nextElement()).elementAt(0);
	    minx = maxx = p1.x;
	    miny = maxy = p1.y;
	    for (Enumeration e = p.strokes(); e.hasMoreElements(); )
	    {
		Vector v = (Vector) e.nextElement();
		for (int i = 0; i < v.size(); i++)
		{
		    Point p2 = (Point) v.elementAt(i);
		    if (p2.x > maxx)
			maxx = p2.x;
		    if (p2.y > maxy)
			maxy = p2.y;
		    if (p2.x < minx)
			minx = p2.x;
		    if (p2.y < miny)
			miny = p2.y;
		}
	    }
	    return Math.max(maxx - minx, maxy - miny);
	}

    /**
     * Calculates the distance between the current pattern and a stored sample
     * @param sampleIndex The index of the stored pattern
     * @return The distance
     */
    private int distance(int sampleIndex, int step)
	{
	    currentModelPoint = filter(step, samples[sampleIndex]);
	    currentModelSlope = filter(step, 
				       sampleSlopeAt[getBank()][sampleIndex]);
	    newLookup(currentUnknownPoint.length, currentModelPoint.length);
	    return D(currentUnknownPoint.length - 1, 
		     currentModelPoint.length - 1) / currentModelPoint.length;
	}

    private byte[][] filter(int step, byte[][] array)
	{
	    if (step <= 1)
		return array;
	    int pos = 0;
	    byte[][] na = new byte[array.length][];
	    for (int i = 0; i < array.length; i++)
		if (i % step == 0)
		    na[pos++] = array[i];
	    byte[][] na2 = new byte[pos][];
	    for (int i = 0; i < pos; i++)
		na2[i] = na[i];
	    return na2;
	}

    private float[] filter(int step, float[] array)
	{
	    if (step <= 1)
		return array;
	    int pos = 0;
	    float[] na = new float[array.length];
	    for (int i = 0; i < array.length; i++)
		if (i % step == 0)
		    na[pos++] = array[i];
	    float [] na2 = new float[pos];
	    for (int i = 0; i < pos; i++)
		na2[i] = na[i];
	    return na2;
	}
		
    private byte[][] cleanStroke(Pattern pat)
	{
	    Vector v = new Vector();
	    for (Enumeration en = pat.strokes(); en.hasMoreElements(); )
	    {
		Vector stroke = (Vector) en.nextElement();
		for (int i = 0; i < stroke.size(); i++)
		    v.addElement(stroke.elementAt(i));
	    }

	    byte[][] res = new byte[v.size()][2];
	    for (int i = 0; i < v.size(); i++)
	    {
		Point p = (Point) v.elementAt(i);
		res[i][0] = (byte) p.x; 
		res[i][1] = (byte) p.y;
	    }
	    return res;
	}

    /**
     * The actual Recursive Elastic Matching function
     */
    private int D(int i, int j)
	{
	    if (lookup[i][j] != -1)
		return lookup[i][j];
	    if (j == 1)
		return lookup[i][j] = pointDistance(i, 1)
		    + Math.min(D(i-1, 1), D(i-1, 0));
	    else
		return lookup[i][j] = pointDistance(i,j)
		    + Math.min(D(i-1, j), Math.min(D(i-1, j-1), D(i-1, j-2)));
	}

    /**
     * Creates a new lookup table for temporary results in
     * the recursion
     * @param n Number of points in the unknown
     * @param m Number of points in the model
     */
    private void newLookup(int n, int m)
	{
	    lookup = new int[n][m];
	    for (int i = 0; i < n; i++)
		for (int j = 0; j < m; j++)
		    lookup[i][j] = -1;
	    lookup[0][0] = pointDistance(0, 0);
	    for (int j = 1; j < m; j++)
		lookup[0][j] = lookup[0][j-1] + pointDistance(0, j);
	    for (int i = 1; i < n; i++)
		lookup[i][0] = lookup[i-1][0] + pointDistance(i, 0);
	}

    /**
     * The distance function between two points
     * @param i Index in pat
     * @param j Index in model
     */
    private int pointDistance(int i, int j)
	{
	    int p1x = currentUnknownPoint[i][0];
	    int p1y = currentUnknownPoint[i][1];
	    int p2x = currentModelPoint[j][0];
	    int p2y = currentModelPoint[j][1];
	    return (p1x-p2x)*(p1x-p2x) + (p1y-p2y)*(p1y-p2y) + 
		slopeDistance(currentUnknownSlope[i], 
			      currentModelSlope[j]);
	}

    /**
     * Initiates the preprocessor.
     */
    private void initPreprocessor()
	{
	    NormFilter nf = new NormFilter(100, 100);
	    SmoothFilter sf = new SmoothFilter();
	    PointDistanceNormFilter psnf = 
		new PointDistanceNormFilter(THRESHOLD);
	    Vector v = new Vector();
	    v.addElement(nf);
	    v.addElement(sf);
	    v.addElement(psnf);
	    preprocessor = new FilterCollection(v);
	}

    /**
     * Loads stored patterns from files
     */
    private void initSamples()
	{
	    // Initiates slopes for all samples
	    sampleSlopeAt = new float[NUM_BANKS][][];
	    
	    for (int i = 0; i < NUM_BANKS; i++)
	    {
		int length = 0;
		byte[][][] sam = null;
		switch (i)
		{
		    case 0:
		    sam = Samples.getNatural();
		    length = sam.length;
		    break;
		    case 1:
		    sam = Samples.getNumeral();
		    length = sam.length;		    
		    break;
		    case 2:		    
		    sam = Samples.getExtended();
		    length = sam.length;		    
		    break;
		};
		
		sampleSlopeAt[i] = new float[length][];
		for (int j = 0; j < length; j++)
		    sampleSlopeAt[i][j] = calcSlopes(sam[j]); 
	    }
	}

    /**
     * Calculates the slope angles for a stroke.
     * @param points The stroke given as a vector of Points
     * @return A Vector of Double containing the slope angles
     */
    private float[] calcSlopes(byte[][] points)
	{
	    float[] angles = new float[points.length];
	    if (points.length <= 0)
		return null;
	    else
	    if (points.length == 1)
	    {
		angles[0] = 0;
		return angles;
	    }

	    for (int i = 0; i < points.length-1; i++)
	    {
		Point p0 = new Point(points[i][0], points[i][1]);
		Point p1 = new Point(points[i+1][0], points[i+1][1]);
		angles[i] = (float) pointSlope(p1, p0);
	    }
	    angles[points.length-1] = angles[points.length-2];
	    return angles;
	}

    private double pointSlope(Point p1, Point p0)
	{
	    double hyp = Math.sqrt((p1.x-p0.x)*(p1.x-p0.x) + 
				   (p1.y-p0.y)*(p1.y-p0.y));
	    double alpha = Math.acos((p1.x-p0.x)/hyp);
	    if (p1.y < p0.y) 
		alpha = 2 * Math.PI - alpha;
	    return alpha;
	}

    private int slopeDistance(float alpha1, float alpha2)
	{
	    double delta = Math.abs(alpha1 - alpha2);
	    return (int) (C * Math.min(delta, 2 * Math.PI - delta));
	}

    /**
     * Gives an index vector of candidates
     */
    private Vector candidates()
	{
	    Vector v = new Vector(N_CANDIDATES);
	    if (N_CANDIDATES >= samples.length)
		{
		    for (int i = 0; i < samples.length; i++)
			v.addElement(new Integer(i));
		    return v;
		}

	    DistIndex[] di = new DistIndex[samples.length];

	    for (int i = 0; i < samples.length; i++)
		di[i] = new DistIndex(distance(i, QUICK_SAMPLE_RATE), i);
	    
	    /* Sorts the array */ 
	    quickSort(di, 0, samples.length - 1);

	    /* Picks the N_CANDIDATES best choices */
	    for (int i = 0; i < N_CANDIDATES; i++)
		v.addElement(new Integer(di[i].index));
	    
	    return v;
	}
    
    private void quickSort(DistIndex[] di, int low, int high)
	{
	    int lo = low;
	    int hi = high;
	    int mid;

	    if (high > low)
	    {
		/* Pick a partitioning element */
		mid = di[ (low + high)/2 ].dist;
		
		// loop through the array until indices cross
		while ( lo <= hi )
		{
		    /* find the first element that is greater than or equal to
		     * the partition element starting from the left Index
		     */
		    while ((lo < high) && (di[lo].dist < mid))
			lo++;
		    
		    /* find an element that is smaller than or equal to
		     * the partition element starting from the right Index
		     */
		    while ((hi > low) && (di[hi].dist > mid))
			hi--;

		    /* swap if indexes have not crossed */
		    if (lo <= hi)
		    {
			swap(di, lo, hi);
			lo++;
			hi--;
		    }
		}

		/* If the right index has not reached the left side of array
		 * must now sort the left partition
		 */
		if (low < hi)
		    quickSort(di, low, hi);

                /* If the left index has not reached the right side of array
		 * must now sort the right partition
		 */
		if (lo < high)
		    quickSort(di, lo, high);
	    }
	}

    private void swap(DistIndex a[], int i, int j)
	{
	    DistIndex t = a[i];
	    a[i] = a[j];
	    a[j] = t;
	}


    private byte[][][] samples;
    private float[][][] sampleSlopeAt;
    private byte[][] currentModelPoint;
    private float[] currentModelSlope;
    private byte[][] currentUnknownPoint;
    private float[] currentUnknownSlope;
    private FilterCollection preprocessor;
    private int[][] lookup;
}

class DistIndex
{
    public DistIndex(int dist, int index)
	{
	    this.dist = dist;
	    this.index = index;
	}

    public int dist;
    public int index;
}
